unit FreeOTFEExplorerWebDAV;

interface

uses
  Windows, Classes,
  IdContext,
  IdCustomHTTPServer,
  SDUGeneral,
  SDUWebDAV,
  SDFilesystem;

type
  TRequestedItem = record
    Filename: string;
    IsFile: boolean;
    Size: ULONGLONG;
  end;
  PRequestedItem = ^TRequestedItem;

  TFreeOTFEExplorerWebDAV = class(TSDUWebDAV)
  protected
    FReturnGarbage: boolean;
    FRequestedItems: TStringList;

    procedure SetReturnGarbage(newValue: boolean);

    procedure Overwrite(pItem: PRequestedItem);

    procedure AddRequestedItemToList(const filename: string; const isFile: boolean; itemSize: ULONGLONG);
    function  GetRequestedItemFromList(const filename: string): PRequestedItem;

    function GetPropFindXMLSingle_FSGetItem(fileOrDir: string; var item: TSDDirItem): boolean; override;
    function GetPropFindXMLSingle_FSGetItem_Dummy(fileOrDir: string; var item: TSDDirItem): boolean;
    function GetPropFindXMLSingle_Dummy(host: string; fileOrDir: string): string;

    // Overridden inherited...
    procedure DoCommandGet(AContext: TIdContext;
      ARequestInfo: TIdHTTPRequestInfo; AResponseInfo: TIdHTTPResponseInfo); override;
    procedure DoCommandDELETE(AContext: TIdContext;
      ARequestInfo: TIdHTTPRequestInfo; AResponseInfo: TIdHTTPResponseInfo); override;
    procedure DoCommandMOVE(AContext: TIdContext;
      ARequestInfo: TIdHTTPRequestInfo; AResponseInfo: TIdHTTPResponseInfo); override;
    procedure DoCommandPUT(AContext: TIdContext;
      ARequestInfo: TIdHTTPRequestInfo; AResponseInfo: TIdHTTPResponseInfo); override;
    procedure DoCommandPROPFIND(AContext: TIdContext;
      ARequestInfo: TIdHTTPRequestInfo; AResponseInfo: TIdHTTPResponseInfo); override;
    procedure DoCommandPROPPATCH(AContext: TIdContext;
      ARequestInfo: TIdHTTPRequestInfo; AResponseInfo: TIdHTTPResponseInfo); override;

  public
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy(); override;

    property ReturnGarbage: boolean read FReturnGarbage write SetReturnGarbage;
    property RequestedItems: TStringList read FRequestedItems;

    procedure ClearDownRequestedItemsList();
  end;

implementation

uses
  SysUtils,
  Shredder,
  SDUHTTPServer,
  SDUClasses;

{$IFDEF _NEVER_DEFINED}
// This is just a dummy const to fool dxGetText when extracting message
// information
// This const is never used; it's #ifdef'd out - SDUCRLF in the code refers to
// picks up SDUGeneral.SDUCRLF
const
  SDUCRLF = ''#13#10;
{$ENDIF}


const
  SECTOR_SIZE = 512;

constructor TFreeOTFEExplorerWebDAV.Create(AOwner: TComponent);
begin
  inherited;

  ReturnGarbage := FALSE;

  FRequestedItems := TStringList.Create();
  FRequestedItems.Sorted := TRUE;

end;

destructor TFreeOTFEExplorerWebDAV.Destroy();
begin
  ClearDownRequestedItemsList();

  FRequestedItems.Free();

  inherited;
end;

function TFreeOTFEExplorerWebDAV.GetRequestedItemFromList(const filename: string): PRequestedItem;
var
  retval: PRequestedItem;
  idx: integer;
  useFilename: string;
begin
  retval := nil;

  useFilename := filename;
  if not(FileSystem.CaseSensitive) then
    begin
    useFilename := uppercase(useFilename);
    end;

  idx := FRequestedItems.IndexOf(useFilename);
  if (idx >= 0) then
    begin
    retval := PRequestedItem(FRequestedItems.Objects[idx]);
    end;

  Result := retval;
end;

procedure TFreeOTFEExplorerWebDAV.AddRequestedItemToList(const filename: string; const isFile: boolean; itemSize: ULONGLONG);
var
  pItem: PRequestedItem;
begin
  pItem := GetRequestedItemFromList(filename);

  if (pItem = nil) then
    begin
    new(pItem);

    // Note: This IS case sensitive - and should be as WebDAV is
    FRequestedItems.AddObject(filename, TObject(pItem));
    end;

  pItem.Filename := filename;
  pItem.IsFile := isFile;
  pItem.Size := itemSize;

end;

procedure TFreeOTFEExplorerWebDAV.ClearDownRequestedItemsList();
var
  pItem: PRequestedItem;
  i: integer;
begin
  // Destroy all associated objects
  for i:=0 to (FRequestedItems.count - 1) do
    begin
    pItem := PRequestedItem(FRequestedItems.Objects[i]);
    Overwrite(pItem);
    Dispose(pItem);
    end;

  Shredder.Overwrite(FRequestedItems);

  FRequestedItems.Clear();

end;

procedure TFreeOTFEExplorerWebDAV.SetReturnGarbage(newValue: boolean);
begin
  if newValue then
    begin
    if (Filesystem <> nil) then
      begin
      Filesystem.ReadOnly := TRUE;
      end;
    end;

  FReturnGarbage := newValue;
end;

procedure TFreeOTFEExplorerWebDAV.Overwrite(pItem: PRequestedItem);
begin
  Shredder.Overwrite(pItem.Filename);
  pItem.IsFile := FALSE;
  pItem.Size := 0;
end;

procedure TFreeOTFEExplorerWebDAV.DoCommandGet(AContext: TIdContext;
      ARequestInfo: TIdHTTPRequestInfo; AResponseInfo: TIdHTTPResponseInfo);
var
  garbageContent: TSDUMemoryStream;
  useSize: ULONGLONG;
  pItem: PRequestedItem;
  localFilename: string;
//  ptrMem: Pointer;
  ptrByte: PByte;
begin
  localFilename:= HTTPReqDocToLocalFile(ARequestInfo.Document);

  if (
      not(ReturnGarbage) or
      (ARequestInfo.Command <> HTTPCMD_GET)
     ) then
    begin
    inherited;

    if (AResponseInfo.ResponseNo = HTTPSTATUS_OK) then
      begin
      if (AResponseInfo.ContentStream <> nil) then
        begin
        AddRequestedItemToList(
                               localFilename,
                               FileSystem.FileExists(localFilename),
                               AResponseInfo.ContentStream.Size
                              );
        end;
      end;
    end
  else
    begin
    useSize := 0;
    pItem := GetRequestedItemFromList(localFilename);
    if (pItem <> nil) then
      begin
      useSize := pItem.Size;
      end;

    // Round up to nearest SECTOR_SIZE byte boundry
    // Note: This sets useSize to SECTOR_SIZE at the minimum - not zero; that's
    //       OK though, and probably better than setting it to 0
    useSize := useSize + (ULONGLONG(SECTOR_SIZE) - ULONGLONG((useSize mod SECTOR_SIZE)));

    garbageContent:= TSDUMemoryStream.Create();
    garbageContent.SetSize(useSize);
    if (useSize > 0) then
      begin
      garbageContent.Position := 0;
      ptrByte := garbageContent.memory;
      FillChar(ptrByte^, useSize, 0);
      end;

    garbageContent.Position := 0;
    AResponseInfo.ContentStream := garbageContent;
    // The response object frees off "fileContent"
    AResponseInfo.FreeContentStream := TRUE;
    AResponseInfo.ResponseNo := HTTPSTATUS_OK;
    end;

end;


function TFreeOTFEExplorerWebDAV.GetPropFindXMLSingle_Dummy(host: string; fileOrDir: string): string;
const
  LOCALE_ENGLISH_UNITED_STATES = 1033;
var
  retval: string;
  item: TSDDirItem;
begin
  retval := '';

  item := TSDDirItem.Create();
  try
    GetPropFindXMLSingle_FSGetItem_Dummy(fileOrDir, item);
    retval := GetPropFindXMLSingle(host, fileOrDir, item);

  finally
    item.Free();
  end;

  Result := retval;
end;

procedure TFreeOTFEExplorerWebDAV.DoCommandPROPFIND(AContext: TIdContext;
  ARequestInfo: TIdHTTPRequestInfo; AResponseInfo: TIdHTTPResponseInfo);
var
  localFilename: string;
  outputXML: string;
begin
  if ReturnGarbage then
    begin
    // Generate generic "OK" reply
    AResponseInfo.ResponseNo := HTTPSTATUS_OK;
    localFilename:= HTTPReqDocToLocalFile(ARequestInfo.Document);

    // lplp - implement functionality so that can take a list of properties interested in

    AddDebugLog('+++localFilename@@:'+localFilename);
    outputXML := GetPropFindXMLSingle_Dummy(ARequestInfo.Host, localFilename);

    outputXML := '<?xml version="1.0"?>'+SDUCRLF+
//x             '<D:multistatus xmlns:D="DAV:">'+SDUCRLF+
           // urn:uuid:c2f41010-65b3-11d1-a29f-00aa00c14882 detailed at: http://xml.coverpages.org/xmlData980105.html and inclued in IIS responses
           '<D:multistatus xmlns:D="DAV:" xmlns:b="urn:uuid:c2f41010-65b3-11d1-a29f-00aa00c14882/">'+SDUCRLF+
           outputXML+SDUCRLF+
           '</D:multistatus>';

    AResponseInfo.ResponseNo := HTTPSTATUS_MULTI_STATUS;
    AResponseInfo.ResponseText := HTTPSTATUS_TEXT_MULTI_STATUS;
    AResponseInfo.ContentType := 'text/xml';
//      outputXML := CompressXML(outputXML);
    AResponseInfo.ContentText := outputXML;
    AddDebugLog('XML:');
    AddDebugLog(outputXML);
    end
  else
    begin
    inherited;
    end;

end;

procedure TFreeOTFEExplorerWebDAV.DoCommandPROPPATCH(AContext: TIdContext;
  ARequestInfo: TIdHTTPRequestInfo; AResponseInfo: TIdHTTPResponseInfo);
begin
  if ReturnGarbage then
    begin
    // Just ignore request, and say it was handled
    AResponseInfo.ResponseNo := HTTPSTATUS_OK;
    end
  else
    begin
    inherited;
    end;

end;

procedure TFreeOTFEExplorerWebDAV.DoCommandPUT(AContext: TIdContext;
  ARequestInfo: TIdHTTPRequestInfo; AResponseInfo: TIdHTTPResponseInfo);
var
  localFilename: string;
  useSize: ULONGLONG;
begin
  if ReturnGarbage then
    begin
    // Just ignore request, and say it was stored
    AResponseInfo.ResponseNo := HTTPSTATUS_OK;
    end
  else
    begin
    inherited;

    if (AResponseInfo.ResponseNo = HTTPSTATUS_OK) then
      begin
      localFilename:= HTTPReqDocToLocalFile(ARequestInfo.Document);

      // Sanity check - if poststream is nil, create zero length stream and use
      // that instead
      useSize := 0;
      if (ARequestInfo.PostStream <> nil) then
        begin
        useSize := ARequestInfo.PostStream.Size;
        end;

      AddRequestedItemToList(
                             localFilename,
                             TRUE,
                             useSize
                            );
      end;
    end;

end;

procedure TFreeOTFEExplorerWebDAV.DoCommandDELETE(AContext: TIdContext;
      ARequestInfo: TIdHTTPRequestInfo; AResponseInfo: TIdHTTPResponseInfo);
begin
  if ReturnGarbage then
    begin
    // Just ignore request, and say it was deleted
    AResponseInfo.ResponseNo := HTTPSTATUS_OK;
    end
  else
    begin
    inherited;
    end;

end;
procedure TFreeOTFEExplorerWebDAV.DoCommandMOVE(AContext: TIdContext;
      ARequestInfo: TIdHTTPRequestInfo; AResponseInfo: TIdHTTPResponseInfo);
begin
  if ReturnGarbage then
    begin
    // Just ignore request, and say it was deleted
    AResponseInfo.ResponseNo := HTTPSTATUS_OK;
    end
  else
    begin
    inherited;
    end;

end;

function TFreeOTFEExplorerWebDAV.GetPropFindXMLSingle_FSGetItem(fileOrDir: string; var item: TSDDirItem): boolean;
var
  retval: boolean;
begin
  retval := inherited GetPropFindXMLSingle_FSGetItem(fileOrDir, item);

  if ReturnGarbage and retval then
    begin
    item.TimestampLastModified := DateTimeToTimeStamp(now());
    end;

  Result := retval;
end;

function TFreeOTFEExplorerWebDAV.GetPropFindXMLSingle_FSGetItem_Dummy(fileOrDir: string; var item: TSDDirItem): boolean;
begin
  item.Filename := ExtractFilename(fileOrDir);
  item.Size := SECTOR_SIZE;
  item.IsFile := TRUE;
  item.IsDirectory := FALSE;
  item.TimestampLastModified := DateTimeToTimeStamp(now());
//    item.CreationDate := DateTimeToTimeStamp(now());
  item.IsHidden := FALSE;

  Result := TRUE;
end;

END.

